package geometry;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Polygon;
import java.util.Collection;
import java.util.Iterator;
import java.util.TreeSet;

import common.Compare;


public class Series extends TreeSet<Point> implements Comparable<Series> {
	private static final long serialVersionUID = -424422617230782026L;

	public Series() {}
	
	// copy constructor
	public Series(Series s)
	{
		for (Point p : s) add(new Point(p));
	}


	// gaussian(xmin, xmax, xstep)
	//   xmin : float = min value of x axis
	//   xmax : float = max value of x axis
	//   xstep: float = distance between x points
	// calculates gaussian series between xmin and max with mean as midpoint
	public static Series gaussian(float xmin, float xmax, float xstep)
	{
		double mean = (xmax-xmin)/2;
		double stddev = (xmax-xmin)/0.15;
		double var = Math.pow(stddev, 2);
		
		Series gauss = new Series();
		for (double x = xmin; Compare.less(x, xmax); x += xstep) {
			double y = Math.exp(-1 * Math.pow(x-mean, 2) / 2*var);
			gauss.add(new Point((float)x, (float)y));
		}
		return gauss;
	}
	
	// offsetBy(base)
	//   base: Series = a series
	// offsets y-values of this series by base series
	public void offsetBy(Series base)
	{
		// find common x-axis resolution for both curves
		TreeSet<Float> x_axis = new TreeSet<Float>();
		for (Point b : base) x_axis.add(b.x);
		for (Point p : this) x_axis.add(p.x); 
		
		// interpolate current and base curve at all axis points
		this.linterp(x_axis);
		base.linterp(x_axis);
		
		// offset current series by base curve y-values
		Series off = this;
		Iterator<Point> oi = off.iterator();
		Iterator<Point> bi = base.iterator();
		while (oi.hasNext() && bi.hasNext()) {
			Point o = oi.next();
			Point b = bi.next();
			if (! Point.fEqual(o.x, b.x)) {
				off.print();
				base.print();
				throw new 
				Error("Series.offsetBy(): base and offset curve not" + 
						"interpolated! o.x=" + o.x + " but b.x=" + b.x);
			}
			o.y += b.y;
		}	
	}


	// normalize(xmin, xmax, ymin, ymax)
	//   xmin: float = minimum x value
	//   xmax: float = maximum x value
	//   ymin: float = minimum y value
	//   ymax: float = maximum y value
	// biases and scales current series to project into given axis ranges
	public void normalize(float xmin, float xmax, float ymin, float ymax)
	{
		float xrange = xmax - xmin;
		float yrange = ymax - ymin;
		for (Point p : this) {
			if (p.x - xmin > xrange) System.out.println("p.x: " + p.x + " xrange: " + xrange);
			p.x = (p.x - xmin) / xrange;

			if (p.y - ymin > yrange) System.out.println("p.y: " + p.y + " yrange: " + yrange);
			p.y = (p.y - ymin) / yrange;
		}
	}

	
	// normalize(box)
	//   box: BBox = a bounding box
	// biases and scales current series to project into given bounding box
	public void normalize(BBox box)
	{
		normalize(box.bl.x, box.tr.x, box.bl.y, box.tr.y);
	}

	
	// scale(xscale, yscale)
	//   xscale: float = x scalar
	//   yscale: float = y scalar
	// scales points in series by given x and y scalars
	public void scale(float xscale, float yscale)
	{
		for (Point p : this)
		{
			p.x *= xscale;
			p.y *= yscale;
		}
	}
	
	// linterp(x)
	// returns y value, linearly interpolated between (x,y) points around x
	public float linterp(float x)
	{
		// find points on either side of x
		Point lo = first();
		Point hi = last();
		Point px = new Point(x, 0);

		// if x is out of bounds, return clamped y-value
		if (x < lo.x) return lo.y;
		if (x > hi.x) return hi.y;

		// otherwise get nearest points surrounding x
		if (floor(px) != null) lo = floor(px);
		if (ceiling(px) != null) hi = ceiling(px);

		// return average of surrounding y-values, weighted by distance
		if (Math.abs(hi.x - lo.x) < Point.epsilon) return lo.y;
		float t = (x - lo.x) / (hi.x - lo.x);
		return (t*hi.y + (1-t)*lo.y);
	}	

	
	// linterp(xs)
	//   xs: Collection<Float> = x values
	// interpolate current series at given x points, adding values to series
	void linterp(Collection<Float> xs)
	{
		for (float x : xs) {
			float y = linterp(x);
			add(new Point(x, y));
		}
	}

	
	// linterp(s)
	//   s: Series = a series
	// interpolate current series for all x points in s, adding values to series
	void linterp(Series s)
	{
		linterp(s.xaxis());
	}


	// xaxis()
	// returns copy of x-axis values
	TreeSet<Float> xaxis()
	{
		TreeSet<Float> xs = new TreeSet<Float>();
		for (Point p : this) {
			xs.add(p.x);
		}
		return xs;
	}

	
	// yaxis()
	// returns copy of y-axis values
	TreeSet<Float> yaxis()
	{
		TreeSet<Float> ys = new TreeSet<Float>();
		for (Point p : this) {
			ys.add(p.y);
		}
		return ys;
	}

	
	// paint(g2d)
	//   g2d: Graphics2D = graphics context
	// paints series as a line plot
	public void paint(Graphics2D g2d)
	{				
		// draw line segments
		Point last = null;
		for (Point p : this) {
			if (last != null) {
				g2d.drawLine((int)last.x, (int)last.y, (int)p.x, (int)p.y);
			}			
			last = p;
		}

		// draw vertices in black
		Paint oldPaint = g2d.getPaint();
		g2d.setPaint(Color.black);
		for (Point p : this) {
			g2d.drawOval((int)p.x, (int)p.y, 1, 1);
		}
		g2d.setPaint(oldPaint);
	}

	// toPolygon()
	// create closed polygon from current series (x,y) points
	public Polygon toPolygon()
	{
		Polygon poly = new Polygon();
		for (Point p : this) {
			poly.addPoint((int)p.x, (int)p.y);
		}
		poly.addPoint((int)last().x, 0);
		return poly;
	}


	// toPolygon(s1, s2)
	//   s1: Series = top series
	//   s2: Series = bottom series, with y-values less than or equal to s1
	// creates polygon from top and bottom curves
	public static Polygon toPolygon(Series s1, Series s2) 
	{
		Polygon poly = new Polygon();
		for (Point p : s1) {
			poly.addPoint((int)p.x, (int)p.y);
		}
		Iterator<Point> pi = s2.descendingIterator();
		while (pi.hasNext()) {
			Point p = pi.next();
			poly.addPoint((int)p.x, (int)p.y);
		}
		return poly;
	}

	
	@Override
	// toString()
	// returns string representation of (x,y) point series
	public String toString()
	{
		String rep = new String("[");
		for (Point p : this) {
			rep += " " + p.toString();
		}
		rep += " ]";
		return rep;
	}
	
	// print()
	// prints text representation of (x,y) points to terminal
	public void print()
	{
		System.out.println(toString());
	}

	// compare series first by start point, then by end point, then by length
	@Override
	public int compareTo(Series s) {
		if (this.equals(s)) return 0;

		int firstComp = first().compareTo(s.first());
		int lastComp = last().compareTo(s.last());
		int lengthComp = Compare.compare(this.size(), s.size());
		int hashComp = Compare.compare(this.hashCode(), s.hashCode());

		if (firstComp != 0) return firstComp;
		if (lastComp != 0) return lastComp;
		if (lengthComp != 0) return lengthComp;
		return hashComp;
	}

}
