package pieChart;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;

import common.Compare;

// Class representing a single wedge in a pie chart
public class PieItem implements Comparable<PieItem> {
	String _name    = "";
	double _percent = 0;
	double _radius  = 0;
	double _theta_start = 0;
	double _theta_end = 0;
	Shape  _shape = null;
	Color  _fill    = Color.black;

	final static double _radians_per_char = 0.15;
	
	PieItem(String name, double percent, double radius, Color color) 
	{
		_name = name;
		_percent = percent;
		_radius = radius;
		_fill = color;
	}

	// copy constructor
	public PieItem(PieItem wedge) 
	{
		_name = new String(wedge._name);
		_percent = wedge._percent;
		_radius = wedge._radius;
		_fill = new Color(wedge._fill.getRed(), wedge._fill.getGreen(),
				wedge._fill.getBlue());
	}
	
	@Override
	public String toString() 
	{
		String rep = _name + ": " + _percent;
		return rep;
	}
	
	@Override
	// equals(o)
	//   o: (PieItem)Object = an Object cast of a PieItem
	// two PieItems are equal iff all of their fields are equal
	public boolean equals(Object o)
	{
		if (o instanceof PieItem) {
			PieItem pi = (PieItem)o;
			return _name == pi._name && _percent == pi._percent &&
			_fill == pi._fill;
		}
		return false;
	}

	
	// compareTo(pi)
	//   pi: PieItem = a pie wedge
	// PieItems are ordered first by color, then percent, then by name
	@Override
	public int compareTo(PieItem pi) 
	{
		int colorComp = Compare.compareHSB(_fill, pi._fill);
		int percComp = Compare.compare(_percent, pi._percent);
		int nameComp = _name.compareTo(pi._name);

		if (colorComp != 0) return colorComp;
		if (percComp != 0) return percComp;
		if (nameComp != 0) return nameComp;
		return 0;
	}

	// layout(start)
	//   start: double = starting angle of wedge sweep, in radians
	// calculates geometry of pie piece given a starting angle
	void layout(double start)
	{
		Polygon tri = new Polygon();
		double x, y;
		double theta = (_percent / 100d) * 2d * Math.PI;
		double diam = 2d * _radius;
		double R = 4d * _radius;

		// save some layout info for later
		_theta_start = start;
		_theta_end = start + theta;
		
		// bounding circle
		Shape circle = new Ellipse2D.Double(-_radius-1, -_radius-1, diam+1, diam+1);
		
		// center of pie point
		tri.addPoint(0, 0);
		
		// left pie edge
		x = R * Math.cos(_theta_start);   y = R * Math.sin(_theta_start);
		tri.addPoint((int)x, (int)y);
		
		// right pie edge
		x = R * Math.cos(_theta_end);   y = R * Math.sin(_theta_end);
		tri.addPoint((int)x, (int)y);

		// calculate wedge by cutting triangle with circle
		Area wedge = new Area(tri);
		wedge.intersect(new Area(circle));
		_shape = wedge;	
	}
	
	// paint(g2d, start)
	//   g2d   : Graphics2D = graphics context
	//   start : double     = starting angle in radians
	//   stroke: boolean    = enable/disable stroke outlines 
	// paints filled pie wedge with black borders, starting at given angle
	// Note: assumes g2d has coordinate axis (0,0) in center of screen
	void paint(Graphics2D g2d, double start, boolean stroke)
	{
		if (_shape == null) layout(start);
		
		// fill pie area with color and stroke black outline
		g2d.setPaint(_fill);
		g2d.fill(_shape);
		g2d.setPaint(Color.black);
		if (stroke) g2d.draw(_shape);
	}
	
	// label(g2d, label, offset)
	//   g2d   : Graphics2D = graphics context
	//   label : String     = label text
	//   offset: double     = offset radius from outermost arc
	// paints label at angle relative to pie piece, with radial offset
	void label(Graphics2D g2d, String label, double offset) 
	{
		// need layout information to draw label
		if (_shape == null) return;

		// rotate text relative to usual (90 degree) orientation
		double theta = (_theta_start + _theta_end) / 2;
		double diff = (Math.PI / 2) - theta;
		diff -= (_radians_per_char * label.length()) - 0.25;

		// draw text along arc by rotations per character
		for (int i = 0; i < label.length(); i++) {
			// Note: assumes origin (0,0) at center of pie
			AffineTransform oldTrans = g2d.getTransform();

			g2d.rotate(-diff);
			g2d.translate(0, _radius+offset);
			g2d.setPaint(Color.black);

			// draw one char at a time to typeset along an arc
			float mat[] = {1, 0, 0, -1, 0, 0};
			g2d.transform(new AffineTransform(mat));
			g2d.drawString(label.substring(i, i+1), 0, 0);

			// advance along arc by certain per-character radian count
			g2d.setTransform(oldTrans);
			if (label.charAt(i) == '.') diff += _radians_per_char/2;
			else if (i+1 < label.length() && label.charAt(i+1) == '.')
				diff += _radians_per_char/1.2;
			else diff += _radians_per_char;
		}
	}
}
