package sawtoothPlot;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;

import common.ColorMap;
import common.Compare;
import geometry.Point;
import geometry.Series;



public class Job extends TreeMap<String, Series> 
                 implements Comparable<Job>, Iterable<Job.Record> {
	ColorMap _map = null;
	String _title = "";
	float _begin = 0;
	float _end = 0;

	// first()
	// return first point in skill history for this job
	public Point first() 
	{
		return this.firstEntry().getValue().first();
	}
	
	// record class for iterating through elements of Job
	public class Record implements Comparable<Record> {
		String _title;
		String _skill;
		Series _data;
		Color  _color;

		Record(String title, String skill, Series data, Color color)
		{
			_title = title;
			_skill = skill;
			_data = data;
			_color = color;
		}

		@Override
		public String toString()
		{
			String rep = "JobRecord { ";
			rep += _skill + " " + _data;
			rep += " }";
			return rep;
		}

		// organize records first by color, then skill, then data
		@Override
		public int compareTo(Record rec) {
			if (this.equals(rec)) return 0;
			
			int colorComp = Compare.compareHSB(_color, rec._color);
			int skillComp = _skill.compareTo(rec._skill);
			int dataComp = _data.compareTo(rec._data);

			if (colorComp != 0) return colorComp;
			if (skillComp != 0) return skillComp;
			return dataComp;
		}
	}

	public class DisplayPolygon {
		Shape _shape;
		Color _color;
	}
	
	// Job(title, begin, end)
	//   title: String = job title
	//   begin: float  = start date
	//   end  : float  = end date
	// create new Job record with given start and end date
	public Job(String title, float begin, float end) 
	{
		_title = title;
		_begin = begin;
		_end = end;
	}

	// add(skill, xp)
	//   skill : String = name of skill
	//   xp    : float  = experience gained
	// add skill name and experience to job record
	void add(String skill, float xp) 
	{
		if (containsKey(skill)) {
			System.out.println("Job: add(): duplicate skill '" + skill + "'");
			return;
		}

		// blank names treated as spacers
		if (skill == "") {
			Series history = new Series();
			history.add(new Point(_begin, 0));
			history.add(new Point(_end, 0));
			put(skill, history);
			return;
		}
		
		// update job with skill learned over course of job
		Series history = new Series();
		
		// begin and end points are linear with time duration
		history.add(new Point(_begin, 0));
		history.add(new Point(_end, xp));
		Series linear = new Series(history);
		
		// middle points are gaussian inspired noise
		double points_per_year = 50;
		double years = _end - _begin;
		double midpoints = years * points_per_year;

		// define noise distribution
		int seed = skill.hashCode();// * (int)Math.floor(xp*10f);
		Random rand = new Random(seed);
		double ymax = xp, ynoise = 0;
		double noise_size = 1.3 * (years / points_per_year), 
		damp_length = midpoints * 0.1;
		if (damp_length < 1) damp_length = 1;
		double damp_size = (1/damp_length) * noise_size;

		if (true) {
		for (int i = 0; i < midpoints; i++) {
			double x = _begin + (i / midpoints) * years;
			double y = linear.linterp((float)x);

			// generate a spike in noise, and add to linear interpolation
			if (Math.abs(ynoise) <= damp_size) {
				ynoise = rand.nextGaussian() * noise_size;
			}
			y += ynoise;			
			history.add(new Point((float)x, (float)y));

			// slowly damp out noise
			if (ynoise > 0) ynoise -= damp_size;
			else ynoise += damp_size;
		}
		}
		
		put(skill, history);
	}

	
	// add(skills)
	//   skills: String... = list of skill names 
	// add skills to job record, all taking equal experience of job duration
	void add(String... skills)
	{
		for (String skill : skills) 
			add(skill, (_end-_begin)/skills.length);
	}

	
	// order Jobs by start date, end date, and skillset size
	@Override
	public int compareTo(Job arg0) {
		// be consistent with equals
		if (this.equals(arg0)) return 0;
		
		// first order by start date
		if (Point.fLess(_begin, arg0._begin)) return -1;
		if (Point.fGreater(_begin, arg0._begin)) return 1;

		// then order by end date
		if (Point.fLess(_end, arg0._begin)) return -1;
		if (Point.fGreater(_end, arg0._begin)) return 1;

		// then order by series length
		if (this.size() < arg0.size()) return -1;
		if (this.size() > arg0.size()) return 1;

		// finally order by series hashcode
		if (this.hashCode() < arg0.hashCode()) return -1;
		else return 1;
	}

	// access record view by iterator
	@Override
	public Iterator<Record> iterator() {		
		return records().iterator();
	}

	// access series ordered by skill color hue
	public Collection<Series> series() {
		ArrayList<Series> series = new ArrayList<Series>();
		for (Record rec : records()) series.add(rec._data);
		return series;
	}

	
	// records()
	// return JobRecord view of series, organized by color (hue)
	public Collection<Record> records()
	{
		TreeSet<Record> byHue = new TreeSet<Record>();
		for (Entry<String, Series> kv : this.entrySet()) {
			String skill  = kv.getKey();			
			Series data   = kv.getValue();
			Color  color  = _map.colorOf(skill);
			byHue.add(new Record(_title, skill, data, color));
		}
		return byHue;
	}

	// paint(g2d)
	//   g2d: Graphics2D = graphics context
	// draw job as series of overarching filled areas
	void paint(Graphics2D g2d)
	{
		for (DisplayPolygon poly : mergedPolygons()) {
		  g2d.setPaint(poly._color);
		  g2d.fill(poly._shape);
		  g2d.setPaint(Color.black);
		  g2d.draw(poly._shape);
		}
	}

	// mergedPolygons()
	// returns list of area polygons merged into contiguously colored regions
	public Collection<DisplayPolygon> mergedPolygons() {
		ArrayList<DisplayPolygon> polygons = new ArrayList<DisplayPolygon>();
		DisplayPolygon last = null;
		for (DisplayPolygon poly : polygons()) {
			if (last == null) {
				last = poly;
				continue;
			}

			// if this color same as last, merge into larger polygon
			if (last._color.equals(poly._color)) {
				Area area = new Area(last._shape);
				area.add(new Area(poly._shape));
				DisplayPolygon merged = new DisplayPolygon();
				merged._color = poly._color;
				merged._shape = area;
				last = merged;
				continue;
			}
			
			// otherwise keep last polygon as-is
			else polygons.add(last);
			
			last = poly;
		}
		polygons.add(last);
		
		return polygons;
	}
	
	// polygons()
	// generate colored polygons representing overarching area series
	public Collection<DisplayPolygon> polygons()
	{
		ArrayList<DisplayPolygon> polygons = new ArrayList<DisplayPolygon>();
		Series last = null;
		for (Record rec : records()) {
			DisplayPolygon poly = new DisplayPolygon();
			poly._color = rec._color;

			if (last == null) poly._shape = rec._data.toPolygon();
			else poly._shape = Series.toPolygon(rec._data, last);
				
			polygons.add(poly);
			last = rec._data;
		}			
		return polygons;
	}
	
	@Override
	public String toString() 
	{
		String rep = "Job {\n";
		for (Record rec : records()) rep += "  " + rec.toString() + "\n";
		rep += "}";
		return rep;
	}
	
	void print()
	{
		System.out.println(toString());
	}
}
