package sawtoothPlot;

import geometry.BBox;
import geometry.Point;
import geometry.Series;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.TreeSet;

import common.ColorMap;


public class JobDB extends TreeSet<Job> {
	
	ColorMap _map = null;     // current skill color map
	float _begin = 99999;     // start date of earliest job, in decimal year
	float _end = 0;           // end date of latest job, in decimal year
	
	JobDB() {}

	JobDB(ColorMap map) 
	{
		_map = map;
	}

	
	// add(title, begin, end, skills...)
	//   title: String = name of job title
	//   begin: float  = start of job duration in decimal year
	//   end  : float  = end of job duration in decimal year
	//   skills...: String... = list of skills practiced during job
	// adds job entry to database
	void add(String title, float begin, float end, String... skills)
	{
		Job job = new Job(title, begin, end);
		job._map = _map;
		job.add(skills);
		add(job);

		if (begin < _begin) _begin = begin;
		if (end > _end) _end = end;
	}

	
	// add(title, year, skills...)
	//   title : String    = job title
	//   year  : float     = year skills are learned in
	//   skills: String... = list of skill names
	// adds skill/year points to database
	void add(String title, float year, String... skills) 
	{
		add(title, year, year+1, skills);
	}


	// getColor(skill)
	//   skill: String = name of skill
	// gets color of named skill in current color map
	Color getColor(String skill)
	{
		return _map.colorOf(skill);
	}
		
	// bbox()
	// calculates smallest bounding box that will contain all current series
	BBox bbox() 
	{
		Point min = new Point(Float.MAX_VALUE, Float.MAX_VALUE);
		Point max = new Point(Float.MIN_VALUE, Float.MIN_VALUE);
		for (Point p : points()) {
			if (p.x < min.x) min.x = p.x;
			if (p.y < min.y) min.y = p.y;

			if (p.x > max.x) max.x = p.x;
			if (p.y > max.y) max.y = p.y;
		}

		if (min.y < 0) min.y = 0;
		return new BBox(min, max);
	}

	// offset()
	// for each job, create overlapping series that describe job
	void offset() 
	{
		Series last = null;
		for (Job job : this) {
			for (Series cur : job.series()) {
				if (last == null) {
					last = cur;
					continue;
				}
				cur.offsetBy(last);
				last = cur;
			}
			last = null;
		}
	}
	
	// points()
	// returns ordered list of all points in all series
	ArrayList<Point> points()
	{
		ArrayList<Point> points = new ArrayList<Point>();
		for (Series series : series()) {
			for (Point point : series) points.add(point);
		}
		
		return points;
	}

	// series()
	// returns list of all series, ordered first by job, then by hue
	ArrayList<Series> series()
	{
		ArrayList<Series> series = new ArrayList<Series>();
		for (Job job : this) {
			for (Series s : job.series()) series.add(s);
		}

		return series;
	}

	// get(skill)
	//   skill: String = skill name
	// returns first data series in database with given skill name
	Series get(String skill)
	{
		Series data = null;
		for (Job job : this) {
			if (! job.containsKey(skill)) continue;
			data = job.get(skill);
			break;
		}
		return data;
	}
	
	// normalize()
	// normalizes all the series relative to their calculated bounding box
	void normalize()
	{
		BBox box = bbox();
		//System.out.println("normalize(): bbox is " + box);
		for (Series s : series()) s.normalize(box);
	}
	
	// scale(sx, sy)
	//   sx: float = x axis scalar
	//   sy: float = y axis scalar
	// scale all series by given x and y axis scalars
	void scale(float sx, float sy)
	{
		for (Point p : points()) {
			p.x *= sx;
			p.y *= sy;
		}
	}

	
	// layout(width, height)
	//   width : int = width of screen in pixels
	//   height: int = height of screen in pixels
	// lays out stacked graph series for a fixed display size
	void layout(int width, int height)
	{
		normalize();
		offset();
		normalize();
		scale(width, height);
	}
	
	void paint(Graphics2D g2d)
	{
		for (Job job : this) {
			job.paint(g2d);
		}
	}

	void print()
	{
		for (Job job : this) System.out.println(job.toString());
	}
}
