package stackedPlot;

import geometry.BBox;
import geometry.Point;
import geometry.Series;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import common.Compare;
import common.DefaultColorMap;


public class SkillDB {

	protected Map<String, Series> _skills = 
		new TreeMap<String, Series>();
	protected Map<String, Color> _colors =
		new TreeMap<String, Color>();
	protected TreeSet<SkillPair> _order = 
		new TreeSet<SkillPair>();
	
	SkillDB() {}

	public class SkillPair implements Comparable<SkillPair> {
		public String _name;
		public float _date;

		SkillPair(String name, float date)
		{
			_name = name;
			_date = date;
		}

		@Override 
		// equality only if date and name are equal
		public boolean equals(Object obj)
		{
			if (obj instanceof SkillPair) {
				SkillPair sobj = (SkillPair)obj;
				return Point.fEqual(sobj._date, this._date) 
				&& sobj._name.equals(this._name);
			}
			return false;
		}

		// order first by date, then by lexicographic ordering
		@Override
		public int compareTo(SkillPair pr) {
			int dateEql = Compare.compare(this._date, pr._date);
			if (dateEql != 0) return dateEql;
			else return this._name.compareTo(pr._name);
		}
	}
	
	// skillNames()
	// returns iterator over all skill names in database, ordered by date
	ArrayList<String> skillNames() 
	{
		ArrayList<String> list = new ArrayList<String>();
		for (SkillPair pr : _order) {
			list.add(pr._name);
		}
		return list;
	}

	// skillValues()
	// returns iterator over all skill values in database, ordered by date
	ArrayList<Series> skillValues()
	{
		ArrayList<Series> list = new ArrayList<Series>();
		for (String skill : skillNames()) {
			list.add(get(skill));
		}
		return list;
	}
	
	// add(year, skills...)
	//   year  : float     = year skills are learned in
	//   skills: String... = list of skill names
	// adds skill/year points to database
	void add(float year, String... skills) 
	{
		add(year-1, year+1, skills);
	}

	// add(begin, end, skills...)
	// same as above, but specified begin and end year, instead of
	// year-1, year+1 as begin and end
	void add(float begin, float end, String... skills)
	{
		for (String skill : skills) {

			// update skill if already exists
			if (_skills.containsKey(skill)) {
				Series series = _skills.get(skill);
				Point last = series.last();
				series.add(new Point(begin, last.y));
				series.add(new Point(end, last.y+1));
			}
			
			// otherwise create new skill
			else {
				Series series = new Series();
				series.add(new Point(begin, 0));
				series.add(new Point(end, 1));
				_skills.put(skill, series);
				_order.add(new SkillPair(skill, begin));
			}
		}
	}

	void setHue(float hue, String... skills)
	{
		for (String skill : skills) {
			if (_skills.containsKey(skill))
				_colors.put(skill, Color.getHSBColor(hue, 0.7f, 0.9f));
		}
	}

	void setGray(float bright, String... skills) 
	{
		for (String skill : skills) {
			if (_skills.containsKey(skill))
				_colors.put(skill, Color.getHSBColor(0, 0, bright));
		}		
	}
	
	Color getColor(String skill)
	{
		return DefaultColorMap.colorMap(skill);
	}
	
	Series get(String skill)
	{
		return _skills.get(skill);
	}
	
	// bbox()
	// calculates smallest bounding box that will contain all current series
	BBox bbox() 
	{
		Point min = new Point(Float.MAX_VALUE, Float.MAX_VALUE);
		Point max = new Point(Float.MIN_VALUE, Float.MIN_VALUE);
		for (Point p : points()) {
			if (p.compareTo(min) < 0) min = p;
			if (p.compareTo(max) > 0) max = p;
		}
		return new BBox(min, max);
	}

	// offset()
	// for each skill series curve, offset y-value of curve by last curve
	void offset() 
	{
		Series last = null;
		for (String skill : skillNames()) {
			Series cur = get(skill);
			if (last == null) {
				last = cur;
				continue;
			}
			cur.offsetBy(last);
			last = cur;
		}
	}
	
	// points()
	// returns ordered list of all points in all series
	ArrayList<Point> points()
	{
		ArrayList<Point> ps = new ArrayList<Point>();
		for (Series series : _skills.values()) {
			for (Point p : series) {
				ps.add(p);
			}
		}
		return ps;
	}
	
	// normalize()
	// normalizes all the series relative to their calculated bounding box
	void normalize()
	{
		// normalize series to project into overall bounding box
		BBox box = bbox();
		for (Series s : this._skills.values()) s.normalize(box);
		
		// extend all series at the edges so that they are defined on x = [0,1]
		for (Series s : this._skills.values()) {
			Point least = new Point(s.first()); 
			least.x = 0;
			Point greatest = new Point(s.last());
			greatest.x = 1;
			s.add(least);
			s.add(greatest);
		}
	}
	
	// scale(sx, sy)
	//   sx: float = x axis scalar
	//   sy: float = y axis scalar
	// scale all series by given x and y axis scalars
	void scale(float sx, float sy)
	{
		for (Point p : points()) {
			p.x *= sx;
			p.y *= sy;
		}
	}

	// print(skills...)
	//   skills...: String... = list of skill names
	// prints description of given skills and set of skill points to terminal
	void print(String... skills)
	{
		for (Map.Entry kv : _skills.entrySet()) {
			if (kv.getKey() instanceof String == false) return;
			
			// if list of skills are given, print only those
			if (skills.length != 0) {
				boolean match = false;
				for (String skill : skills) {
					if (skill == (String)kv.getKey()) match = true;
				}
				if (!match) continue;
			}
			
			// print skill name
			System.out.print((String)kv.getKey());
			System.out.println(":");

			// print skill plot values
			if (kv.getValue() instanceof Series) {
				((Series)kv.getValue()).print();
			}
		}
	}

	
	// paint(g2d)
	//   g2d: Grapics2D = graphics context
	// paints series as multiple overlaid area plots
	void paint(Graphics2D g2d)
	{
		// compute polygons from areas between series, and fill with colors
		Series last = null;
		for (Series cur : skillValues()) {
			if (last == null) {
				Polygon p = cur.toPolygon();
				g2d.fill(p);
				last = cur;
				continue;
			}
			Polygon p = Series.toPolygon(cur, last);
			g2d.fill(p);
		}
	}


	// paint(g2d, skill)
	//   g2d  : Graphics2D = graphics context
	//   skill: String     = skill name
	// paint given skill as closed polygon with borders formed from 
	// given skill series and previous skill series
	void paint(Graphics2D g2d, String skill)
	{
		// find series just before skill series, if present
		String last = null;
		for (String sk : skillNames()) {
			if (sk == skill) break;
			last = sk;
		}

		// if skill is first in series, draw as standalone polygon
		if (last == null) {
			Polygon p = get(skill).toPolygon();
			g2d.fill(p);
			return;
		}

		// otherwise use previous series to define skill polygon
		Polygon p = Series.toPolygon(get(skill), get(last));
		g2d.fill(p);
	}

}
